#![allow(non_upper_case_globals)]
use crate::tarray::{FString, TArray};
use bitflags::bitflags;
use std::ptr::NonNull;

#[derive(Debug)]
#[repr(transparent)]
pub struct VTbl(NonNull<()>);

impl VTbl {}

bitflags! {
    #[repr(C)]
    pub struct EObjectFlags: u32 {
        const RF_NoFlags                      = 0x00000000;
        const RF_Public                       = 0x00000001;
        const RF_Standalone                   = 0x00000002;
        const RF_MarkAsNative                 = 0x00000004;
        const RF_Transactional                = 0x00000008;
        const RF_ClassDefaultObject           = 0x00000010;
        const RF_ArchetypeObject              = 0x00000020;
        const RF_Transient                    = 0x00000040;
        const RF_MarkAsRootSet                = 0x00000080;
        const RF_TagGarbageTemp               = 0x00000100;
        const RF_NeedInitialization           = 0x00000200;
        const RF_NeedLoad                     = 0x00000400;
        const RF_KeepForCooker                = 0x00000800;
        const RF_NeedPostLoad                 = 0x00001000;
        const RF_NeedPostLoadSubobjects       = 0x00002000;
        const RF_NewerVersionExists           = 0x00004000;
        const RF_BeginDestroyed               = 0x00008000;
        const RF_FinishDestroyed              = 0x00010000;
        const RF_BeingRegenerated             = 0x00020000;
        const RF_DefaultSubObject             = 0x00040000;
        const RF_WasLoaded                    = 0x00080000;
        const RF_TextExportTransient          = 0x00100000;
        const RF_LoadCompleted                = 0x00200000;
        const RF_InheritableComponentTemplate = 0x00400000;
        const RF_DuplicateTransient           = 0x00800000;
        const RF_StrongRefOnFrame             = 0x01000000;
        const RF_NonPIEDuplicateTransient     = 0x02000000;
        const RF_Dynamic                      = 0x04000000;
        const RF_WillBeLoaded                 = 0x08000000;
        const RF_HasExternalPackage           = 0x10000000;
        }
}

use crate::global_tables::names::GNAMES;

#[repr(C)]
#[derive(Debug)]
pub struct UObject {
    pub vtbl: VTbl,
    pub object_flags: EObjectFlags, //EObjectFlags,
    pub internal_index: u32,
    pub class: Option<NonNull<UClass>>,
    pub name: FName,
    pub outer: Option<NonNull<UObject>>,
}

#[repr(C)]
#[derive(Debug, Eq, PartialEq, Clone, Copy, Hash)]
pub struct FName {
    pub comparison_index: u32,
    pub number: u32,
}

impl FName {
    pub fn get_name(&self) -> anyhow::Result<String> {
        GNAMES
            .read()
            .unwrap()
            .as_names()
            .unwrap()
            .fname_to_string(self)
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct UField {
    pub uobject: UObject,
    pub next: Option<NonNull<UField>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UEnum {
    pub ufield: UField,
    pub cpp_type: FString,
    pub names: TArray<FName>,
    pub cpp_form: u32,
}

#[repr(C)]
#[derive(Debug)]
pub struct UStruct {
    pub(crate) ufield: UField,
    pub(crate) super_field: Option<NonNull<UStruct>>,
    pub(crate) children: Option<NonNull<UField>>,
    pub(crate) property_size: u32,
    pub(crate) min_alignment: u32,
    padding1: [u8; 0x48],
}

#[repr(C)]
#[derive(Debug)]
pub struct UClass {
    pub(crate) ustruct: UStruct,
    padding1: [u8; 0xF8],
}

#[repr(C)]
#[derive(Debug)]
pub struct UProperty {
    pub(crate) ufield: UField,
    array_dim: i32,
    element_size: i32,
    property_flags: u64,
    rep_index: i16,
    rep_notify_function: FName,
    offset: i32,
    property_link_next: Option<NonNull<UProperty>>,
    next_ref: Option<NonNull<UProperty>>,
    destructor_link_next: Option<NonNull<UProperty>>,
    post_construct_link_next: Option<NonNull<UProperty>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UByteProperty {
    pub(crate) uproperty: UProperty,
    uenum: Option<NonNull<UEnum>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UBoolProperty {
    pub(crate) uproperty: UProperty,
    field_size: u8,
    byte_offset: u8,
    byte_mask: u8,
    field_mask: u8,
}

#[repr(C)]
#[derive(Debug)]
pub struct UObjectProperty {
    pub(crate) uproperty: UProperty,
    property_class: Option<NonNull<UClass>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UClassProperty {
    pub(crate) uobjectproperty: UObjectProperty,
    meta_class: Option<NonNull<UClass>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UAssetClassProperty {
    pub(crate) uobjectproperty: UObjectProperty,
    meta_class: Option<NonNull<UClass>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UInterfaceProperty {
    pub(crate) uproperty: UProperty,
    interface_class: Option<NonNull<UClass>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UStructProperty {
    pub(crate) uproperty: UProperty,
    ustruct: Option<NonNull<UStruct>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UArrayProperty {
    pub(crate) uproperty: UProperty,
    inner: Option<NonNull<UProperty>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UMapProperty {
    pub(crate) uproperty: UProperty,
    key_prop: Option<NonNull<UProperty>>,
    value_prop: Option<NonNull<UProperty>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UDelegateProperty {
    pub(crate) uproperty: UProperty,
    signature_function: Option<NonNull<UFunction>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UMulticastDelegateProperty {
    pub(crate) uproperty: UProperty,
    signature_function: Option<NonNull<UFunction>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UEnumProperty {
    pub(crate) uproperty: UProperty,
    underlying_type: Option<NonNull<UProperty>>,
    uenum: Option<NonNull<UEnum>>,
}

#[repr(C)]
#[derive(Debug)]
pub struct UFunction {
    pub(crate) ustruct: UStruct,
    function_flags: u32,
    rep_offset: u16,
    num_params: u8,
    params_size: u16,
    return_value_offset: u16,
    rpc_id: u16,
    rpc_response_id: u16,
    first_property_to_init: Option<NonNull<UProperty>>,
    function: Option<NonNull<()>>,
}
