use std::{
    cell::RefCell,
    collections::{hash_map::Entry, HashMap},
    sync::Mutex,
};

use anyhow::Context;
use once_cell::sync::{Lazy, OnceCell};

use crate::global_tables::names::GNAMES;

#[repr(C)]
#[derive(Debug, Eq, PartialEq, Clone, Copy, Hash)]
pub struct FName {
    pub comparison_index: u32,
    pub number: u32,
}

impl FName {
    pub fn new(name: &'static str) -> anyhow::Result<Self> {
        static CACHE: Lazy<Mutex<RefCell<HashMap<&'static str, u32>>>> =
            Lazy::new(|| Mutex::new(RefCell::new(HashMap::new())));

        let cache = CACHE.lock().unwrap();

        let name = match cache.borrow_mut().entry(name) {
            Entry::Occupied(entry) => Self {
                comparison_index: *entry.get(),
                number: 0,
            },
            Entry::Vacant(entry) => {
                let name = GNAMES
                    .read()
                    .unwrap()
                    .as_names()
                    .unwrap()
                    .enumerate()
                    .find_map(|(i, entry)| (entry.as_str() == name).then_some(i))
                    .context("could not find fname.")? as u32;

                entry.insert(name);

                Self {
                    comparison_index: name,
                    number: 0,
                }
            }
        };

        Ok(name)
    }

    pub fn none() -> FName {
        static NAME: Lazy<FName> = Lazy::new(|| FName::new("None").expect("FName \"None\"."));

        *NAME
    }

    pub fn get_name(&self) -> anyhow::Result<String> {
        GNAMES
            .read()
            .unwrap()
            .as_names()
            .unwrap()
            .fname_to_string(self)
    }
}
