use std::{cell::UnsafeCell, ops::Index, ptr::NonNull, sync::RwLock};

use crate::v2_types::{self, traits::AsUObject};

lazy_static::lazy_static! {
    pub static ref GOBJECTS: RwLock<GObjects<v2_types::UObject>> = RwLock::new(GObjects::new());
}

#[derive(Debug, Default)]
pub struct GObjects<Object: Sized> {
    objects: Option<NonNull<UnsafeCell<ObjectArray<Object>>>>,
}

impl<Object> GObjects<Object>
where
    Object: Sized,
{
    pub const fn new() -> Self {
        Self { objects: None }
    }

    pub fn set_objects(&mut self, objects: NonNull<ObjectArray<Object>>) {
        self.objects = Some(objects.cast());
    }

    pub fn as_objects(&self) -> Option<&ObjectArray<Object>> {
        self.objects
            .map(|objects| unsafe { &*objects.as_ref().get() })
    }
}

unsafe impl<Object: Sized> Send for GObjects<Object> {}
unsafe impl<Object: Sized> Sync for GObjects<Object> {}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArrayItem<Object: Sized> {
    object: Option<Object>,
    sn: u32,
}

unsafe impl<Object: Sized> Send for ObjectArrayItem<Object> {}
unsafe impl<Object: Sized> Sync for ObjectArrayItem<Object> {}

impl<Object> ObjectArrayItem<Object>
where
    Object: Sized,
{
    pub fn object(&self) -> Option<Object>
    where
        Object: Copy,
    {
        self.object
    }

    pub fn get_object(&self) -> &Option<Object> {
        &self.object
    }

    pub fn serial_number(&self) -> u32 {
        self.sn
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArrayInner<Object: Sized> {
    objects: NonNull<Option<NonNull<ObjectArrayItem<Object>>>>,
    pre_allocated_objects: Option<NonNull<ObjectArrayItem<Object>>>,
    max_elements: u32,
    num_elements: u32,
    max_chunks: u32,
    num_chunks: u32,
}

impl<Object: Sized> ObjectArrayInner<Object> {
    const ELEMENTS_PER_CHUNK: usize = 64 * 1024;

    fn chunks_as_slice(&self) -> &[Option<NonNull<ObjectArrayItem<Object>>>] {
        unsafe { std::slice::from_raw_parts(self.objects.as_ptr(), self.num_chunks as usize) }
    }

    pub fn get_chunks(&self) -> Vec<&[ObjectArrayItem<Object>]> {
        (0..self.num_chunks as usize)
            .filter_map(|i| self.chunk_as_slice(i))
            .collect()
    }

    fn get_chunk_size(&self, chunk_index: usize) -> Option<usize> {
        if chunk_index < self.num_chunks as usize {
            if chunk_index * Self::ELEMENTS_PER_CHUNK <= self.num_elements as usize {
                Some(Self::ELEMENTS_PER_CHUNK)
            } else {
                Some(self.num_elements as usize % Self::ELEMENTS_PER_CHUNK)
            }
        } else {
            None
        }
    }

    fn chunk_as_slice(&self, chunk_index: usize) -> Option<&[ObjectArrayItem<Object>]> {
        let chunks = self.chunks_as_slice();
        let chunk = unsafe {
            chunks[chunk_index].map(|ptr| {
                std::slice::from_raw_parts(ptr.as_ptr(), self.get_chunk_size(chunk_index).unwrap())
            })
        };

        chunk
    }

    /// returns the chunk index and the index inside that chunk, or none if the index is out of range
    fn get_chunked_index(&self, i: usize) -> Option<(usize, usize)> {
        if i < self.num_elements as usize {
            let chunk_index = i / Self::ELEMENTS_PER_CHUNK;
            let within_chunk_index = i % Self::ELEMENTS_PER_CHUNK;

            Some((chunk_index, within_chunk_index))
        } else {
            None
        }
    }

    pub fn get_index(&self, i: usize) -> Option<&ObjectArrayItem<Object>> {
        if let Some((chunk, within_chunk)) = self.get_chunked_index(i) {
            self.chunk_as_slice(chunk)?.get(within_chunk)
        } else {
            None
        }
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArray<Object: Sized> {
    first_gc_index: u32,
    last_non_gc_index: u32,
    max_objects_not_considered_by_gc: u32,
    open_for_disregard_for_gc: bool,
    inner: ObjectArrayInner<Object>,
}

unsafe impl<Object: Sized> Send for ObjectArray<Object> {}
unsafe impl<Object: Sized> Sync for ObjectArray<Object> {}

impl<Object: Sized> ObjectArray<Object> {
    pub fn len(&self) -> usize {
        self.inner.num_elements as usize
    }

    pub fn get(&self, i: usize) -> Option<&ObjectArrayItem<Object>> {
        self.inner.get_index(i)
    }

    pub fn iter(&self) -> iter::ObjectArrayIterator<'_, Object> {
        iter::ObjectArrayIterator::new(self)
    }

    pub fn par_iter(&self) -> par_iter::ObjectArrayChunkedParallelIterator<'_, Object> {
        par_iter::ObjectArrayChunkedParallelIterator::new(self.inner.get_chunks())
    }
}

impl<Object> Index<usize> for ObjectArray<Object> {
    type Output = ObjectArrayItem<Object>;

    fn index(&self, i: usize) -> &Self::Output {
        self.inner.get_index(i).expect("Out of bounds access")
    }
}

pub mod iter {
    use super::{ObjectArray, ObjectArrayItem};

    pub struct ObjectArrayIterator<'a, Object>
    where
        Object: Sized,
    {
        objects: &'a ObjectArray<Object>,
        index: usize,
    }

    impl<'a, Object> ObjectArrayIterator<'a, Object>
    where
        Object: Sized,
    {
        pub(super) fn new(objects: &'a ObjectArray<Object>) -> Self {
            Self { objects, index: 0 }
        }
    }

    impl<'a, Object> Iterator for ObjectArrayIterator<'a, Object>
    where
        Object: Sized,
    {
        type Item = &'a ObjectArrayItem<Object>;

        fn next(&mut self) -> Option<Self::Item> {
            let item = if self.index < self.objects.len() {
                Some(&self.objects[self.index])
            } else {
                None
            };

            self.index += 1;

            item
        }
    }

    impl<'a, Object> IntoIterator for &'a ObjectArray<Object>
    where
        Object: Sized,
    {
        type Item = &'a ObjectArrayItem<Object>;
        type IntoIter = ObjectArrayIterator<'a, Object>;

        fn into_iter(self) -> Self::IntoIter {
            ObjectArrayIterator {
                objects: self,
                index: 0,
            }
        }
    }
}

pub mod par_iter {
    pub use rayon::iter::{IntoParallelIterator, ParallelIterator};

    use super::ObjectArrayItem;

    pub struct ObjectArrayChunkedParallelIterator<'a, Object>
    where
        Object: Sized,
    {
        chunks: Vec<&'a [ObjectArrayItem<Object>]>,
    }

    impl<'a, Object> ObjectArrayChunkedParallelIterator<'a, Object>
    where
        Object: Sized,
    {
        pub fn new(chunks: Vec<&'a [ObjectArrayItem<Object>]>) -> Self {
            Self { chunks }
        }
    }

    impl<'a, Object> ParallelIterator for ObjectArrayChunkedParallelIterator<'a, Object>
    where
        Object: Sized,
    {
        type Item = &'a ObjectArrayItem<Object>;

        fn drive_unindexed<C>(self, consumer: C) -> C::Result
        where
            C: rayon::iter::plumbing::UnindexedConsumer<Self::Item>,
        {
            self.chunks
                .into_par_iter()
                .flatten_iter()
                .drive_unindexed(consumer)
        }
    }
}

pub trait FindObject<Object> {
    fn find_object<S>(&self, object_name: S) -> Option<Object>
    where
        S: Into<String>;
}

pub trait FindClass {
    fn find_class<S>(&self, class_name: S) -> Option<v2_types::UClass>
    where
        S: Into<String>;
}

impl FindClass for ObjectArray<v2_types::UObject> {
    fn find_class<S>(&self, class_name: S) -> Option<v2_types::UClass>
    where
        S: Into<String>,
    {
        use crate::v2_types::traits::UObjectNonConst;
        let class_name = class_name.into();

        self.iter()
            .filter_map(|item| item.object())
            .filter(|object| {
                object
                    .get_full_name()
                    .map(|name| name == class_name)
                    .unwrap_or(false)
            })
            .map(|object| (unsafe { object.cast() }))
            .next()
    }
}
