use std::{
    collections::HashSet,
    io::{BufWriter, Write},
};

use anyhow::Context;
use itertools::Itertools;

use crate::{
    sdk::{
        canonicalize_name, Class, ClassField, ClassMethod, Enum, PrimitiveType, ProcessedPackage,
        Sdk, Type, Types,
    },
    v2_types::{
        traits::{AsUObject, UObjectNonConst, UStructNonConst},
        UObject,
    },
};

pub trait RustType {
    fn rust_type<W: Write>(&self, sdk: &Sdk, w: &mut W) -> anyhow::Result<()>;
}

impl<T> RustType for T
where
    T: AsUObject,
{
    fn rust_type<W: Write>(&self, sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
        sdk.find_type(self.as_uobject())
            .context("could not find type")?
            .rust_type(sdk, w)
    }
}

impl RustType for Types {
    fn rust_type<W: Write>(&self, _sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
        match self {
            Types::Class(class) => {
                write!(w, "{}", class.rust_name())?;
            }
            Types::Enum(enm) => {
                // FIXME: this?
                write!(w, "{}", enm.name)?;
            }
        }

        Ok(())
    }
}

impl RustType for Type {
    fn rust_type<W: Write>(&self, sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
        match self {
            Type::Ptr(ptr) => {
                write!(w, "Option<NonNull<")?;
                ptr.rust_type(sdk, w)?;
                write!(w, ">>")?;
            }
            Type::Ref(_) => todo!(),
            Type::WeakPtr(ptr) => {
                write!(w, "TWeakObjectPtr<")?;
                ptr.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::SoftPtr(ptr) => {
                write!(w, "TSoftObjectPtr<")?;
                ptr.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::LazyPtr(ptr) => {
                write!(w, "TLazyObjectPtr<")?;
                ptr.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::AssetPtr(ptr) => {
                write!(w, "TAssetPtr<")?;
                ptr.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::Array(array) => {
                write!(w, "TArray<")?;
                array.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::Primitive(ty) => {
                write!(
                    w,
                    "{}",
                    match ty {
                        PrimitiveType::Bool => "bool",
                        PrimitiveType::U8 => "u8",
                        PrimitiveType::U16 => "u16",
                        PrimitiveType::U32 => "u32",
                        PrimitiveType::U64 => "u64",
                        PrimitiveType::I8 => "i8",
                        PrimitiveType::I16 => "i16",
                        PrimitiveType::I32 => "i32",
                        PrimitiveType::I64 => "i64",
                        PrimitiveType::F32 => "f32",
                        PrimitiveType::F64 => "f64",
                        PrimitiveType::Custom(custom) => custom,
                    }
                )?;
            }
            Type::RawArray { ty, len } => {
                write!(w, "[")?;
                ty.rust_type(sdk, w)?;
                write!(w, "; {}]", len)?;
            }
            Type::Name => {
                write!(w, "FName")?;
            }
            Type::String => {
                write!(w, "FString")?;
            }
            Type::Text => {
                write!(w, "FText")?;
            }
            Type::Enum { enum_type, .. } => {
                write!(w, "TEnumAsByte<")?;
                enum_type.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::Class(class) => {
                write!(w, "Option<")?;
                class.rust_type(sdk, w)?;
                write!(w, ">")?;
            }
            Type::Struct(class) => {
                class.rust_type(sdk, w)?;
            }
        }
        Ok(())
    }
}

pub fn generate_enum<W: Write>(enm: &Enum, _sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
    let Enum { name, values } = enm;

    writeln!(w, "#[repr(u8)]")?;
    writeln!(
        w,
        "#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]"
    )?;
    writeln!(w, "pub enum {name} {{")?;

    for (i, value) in values.iter().enumerate().take(256) {
        writeln!(w, "\t{value} = {i},")?;
    }
    writeln!(w, "}}")?;

    Ok(())
}

pub fn generate_class<W: Write>(class: &Class, _sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
    let Class {
        is_class,
        size,
        full_name,
        ..
    } = class;

    let name = class.rust_name();

    writeln!(w, "#[repr(transparent)]")?;
    if !is_class {
        writeln!(w, "#[derive(Debug)]")?;
        writeln!(w, "pub struct {name}(pub UnsafeCell<[u8; {size}]>);")?;

        writeln!(w, "unsafe impl Send for {name} {{}}")?;
        writeln!(w, "unsafe impl Sync for {name} {{}}")?;

        write!(
            w,
            r#"
impl Eq for {name} {{}}

impl PartialEq for {name} {{
    fn eq(&self, other: &Self) -> bool {{
        unsafe {{ &*self.0.get() }}.eq(unsafe {{ &*other.0.get() }})
    }}
}}

impl Clone for {name} {{
    fn clone(&self) -> Self {{
        Self(UnsafeCell::new(unsafe {{ &*self.0.get() }}.clone()))
    }}
}}
"#
        )?;

        write!(
            w,
            r#"
impl AsPtr for {name} {{
    fn as_ptr(&self) -> *const u8 {{
        self.0.get().cast()
    }}
    fn as_mut_ptr(&self) -> *mut u8 {{
        self.0.get().cast()
    }}
}}
"#
        )?;

        write!(
            w,
            r#"
impl {name} {{
    pub fn zeroed() -> Self {{
        unsafe {{ core::mem::MaybeUninit::<Self>::zeroed().assume_init() }}
    }}
}}
"#
        )?;
    } else {
        writeln!(w, "#[derive(Debug, Eq, PartialEq, Copy, Clone)]")?;
        writeln!(w, "pub struct {name}(pub NonNull<UnsafeCell<u8>>);")?;

        writeln!(w, "unsafe impl Send for {name} {{}}")?;
        writeln!(w, "unsafe impl Sync for {name} {{}}")?;

        write!(
            w,
            r#"
impl AsUObject for {name} {{
    fn as_uobject(&self) -> UObject {{
        UObject(self.0)
    }}

    fn from_uobject(obj: &UObject) -> Self {{
        Self(obj.0)
    }}
}}
"#
        )?;

        write!(
            w,
            r#"
impl AsPtr for {name} {{
    fn as_ptr(&self) -> *const u8 {{
        unsafe {{ self.0.as_ref().get() as _ }}
    }}
    fn as_mut_ptr(&self) -> *mut u8 {{
        unsafe {{ self.0.as_ref().get() as _ }}
    }}
}}
"#
        )?;

        writeln!(w, "impl StaticClass for {name} {{")?;
        writeln!(w, "fn get_static_class() -> Option<UClass> {{")?;
        write!(w, "let class: Option<UClass> = ")?;
        generate_find_object(&full_name, w)?;
        writeln!(w, ";")?;
        writeln!(w, "class")?;
        writeln!(w, "}}")?;
        writeln!(w, "}}")?;
    }

    Ok(())
}

pub fn generate_class_impl<W: Write>(class: &Class, sdk: &Sdk, w: &mut W) -> anyhow::Result<()> {
    let Class {
        super_class,
        fields,
        methods,
        is_class,
        ..
    } = class;

    let name = class.rust_name();

    writeln!(w, "pub trait {}Fields: AsPtr {{", name)?;
    for field in fields {
        write!(w, "fn {}(&self) -> &", field.name)?;
        field.ty.rust_type(sdk, w)?;
        writeln!(
            w,
            " {{unsafe {{ &*self.as_ptr().offset({}).cast() }} }}",
            field.offset
        )?;

        write!(w, "fn {}_mut(&mut self) -> &mut ", field.name)?;
        field.ty.rust_type(sdk, w)?;
        writeln!(
            w,
            " {{unsafe {{ &mut *self.as_mut_ptr().offset({}).cast() }} }}",
            field.offset
        )?;

        write!(w, "fn set_{0}(&mut self, {0}: ", field.name)?;
        field.ty.rust_type(sdk, w)?;
        write!(w, ") -> ()")?;
        writeln!(
            w,
            " {{*unsafe {{ &mut *self.as_mut_ptr().offset({}).cast() }} = {}; }}",
            field.offset, field.name
        )?;
    }
    writeln!(w, "}}")?;
    writeln!(w, "impl {name}Fields for {name} {{}}")?;

    if *is_class {
        for method in methods {
            generate_method_params(class, method, sdk, w)?;
        }

        writeln!(w, "pub trait {}Methods: AsUObject {{", name)?;

        for method in methods {
            generate_method(class, method, sdk, w)?;
        }
        writeln!(w, "}}")?;
        writeln!(w, "impl {name}Methods for {name} {{}}")?;
    }

    if let Some(supr) = super_class {
        let iter = core::iter::once(*supr).chain(supr.iter_super_structs());
        for parent in iter {
            if let Some(parent) = sdk.find_type(parent.as_uobject()) {
                match parent {
                    Types::Class(super_class) => {
                        writeln!(w, "impl {}Fields for {name} {{}}", super_class.rust_name())?;
                        if *is_class {
                            writeln!(w, "impl {}Methods for {name} {{}}", super_class.rust_name())?;
                        }
                    }
                    _ => {}
                }
            }
        }
    }

    // TODO: functions

    Ok(())
}

fn generate_find_object<W: Write>(name: &str, w: &mut W) -> anyhow::Result<()> {
    write!(
        w,
        r#"
{{
    static OBJECT: OnceCell<Option<UObject>> = OnceCell::new();
    OBJECT.get_or_init(|| {{
        match find_object("{name}") {{
                Some(class) => {{Some(class)}},
                None => {{
                    log::error!("static object {name} not found!");
                    None
                }}
            }}
    }}).map(|object| unsafe {{ object.cast() }})
}}
"#
    )?;
    Ok(())
}

pub fn generate_method<W: Write>(
    class: &Class,
    method: &ClassMethod,
    sdk: &Sdk,
    w: &mut W,
) -> anyhow::Result<()> {
    write!(w, "fn {}(&self", method.name)?;

    for param in &method.parameters {
        match param {
            crate::sdk::ParameterKind::Default(parm) => {
                write!(w, ",{}: ", parm.name)?;
                parm.ty.rust_type(sdk, w)?;
            }
            crate::sdk::ParameterKind::Out(parm) => {
                write!(w, ",{}: &mut ", parm.name)?;
                parm.ty.rust_type(sdk, w)?;
                write!(w, "/*(OutParm)*/")?;
            }
            _ => {}
        }
    }

    write!(w, ") -> ")?;
    match &method.return_type {
        Some(param) => {
            param.ty.rust_type(sdk, w)?;
        }
        None => write!(w, "()")?,
    }

    writeln!(w, " {{")?;

    write!(w, "let mut func: UFunction =")?;
    generate_find_object(&method.full_name, w)?;
    writeln!(w, ".expect(\"function '{}' not found\");", method.full_name)?;

    let params_name = format!("{}_{}_Params", class.rust_name(), method.name);
    writeln!(w, "let mut params = {params_name}::zeroed();")?;

    for param in &method.parameters {
        match param {
            crate::sdk::ParameterKind::Default(param) => {
                writeln!(w, "params.{0} = {0};", param.name)?;
            }
            crate::sdk::ParameterKind::Out(param) => {
                // swap so we dont have to worry about Clone or Copy
                writeln!(w, "core::mem::swap(&mut params.{0}, {0});", param.name)?;
            }
            _ => {}
        }
    }

    writeln!(w, "let flags = *func.function_flags();")?;
    writeln!(w, "process_event(self.as_uobject(), func, &mut params);")?;
    writeln!(w, "func.set_function_flags(flags);")?;

    for param in &method.parameters {
        match param {
            crate::sdk::ParameterKind::Out(param) => {
                // swap back
                writeln!(w, "core::mem::swap({0}, &mut params.{0});", param.name)?;
            }
            _ => {}
        }
    }

    match &method.return_type {
        Some(param) => {
            writeln!(w, "params.{}", param.name)?;
        }
        None => {}
    }

    writeln!(w, "}}")?;

    Ok(())
}

pub fn generate_method_params<W: Write>(
    class: &Class,
    method: &ClassMethod,
    sdk: &Sdk,
    w: &mut W,
) -> anyhow::Result<()> {
    let name = format!("{}_{}_Params", class.rust_name(), method.name);
    write!(w, "#[repr(C)]\n#[derive(Debug)]\npub struct {name} {{\n")?;

    for param in &method.parameters {
        write!(w, "\tpub {}: ", param.as_param().name)?;
        param.as_param().ty.rust_type(sdk, w)?;
        write!(w, ",\n")?;
    }

    write!(w, "}}\n")?;

    write!(
        w,
        r#"
impl {name} {{
    fn zeroed() -> Self {{
        unsafe {{ core::mem::MaybeUninit::<Self>::zeroed().assume_init() }}
    }}
}}
"#
    )?;

    Ok(())
}

pub fn generate_partial_sdk_to_tmp(
    sdk: &Sdk,
    packages: HashSet<&'static str>,
) -> anyhow::Result<()> {
    std::fs::create_dir_all("z:/tmp/ark_sdk/")?;
    let file = std::fs::File::create("z:/tmp/ark_sdk/mod.rs")?;
    let mut writer = BufWriter::new(file);

    writeln!(writer, "use core::cell::UnsafeCell;")?;
    writeln!(writer, "use core::ptr::NonNull;")?;
    writeln!(writer, "use super::AsPtr;")?;
    writeln!(writer, "use super::process_event;")?;
    writeln!(writer, "use super::StaticClass;")?;

    let packages = sdk.packages.iter().filter(|(_, pkg)| {
        let pkg_name = canonicalize_name(
            &pkg.package
                .get_full_name()
                .expect("could not get package name"),
        )
        .to_string();

        packages.contains(pkg_name.as_str())
    });

    let mut processed_packages = HashSet::new();

    for (obj, pkg) in packages {
        if processed_packages.insert(*obj) {
            generate_package(sdk, &mut writer, &mut processed_packages, pkg)?;
        }
    }

    todo!()
}

fn generate_package<W: Write>(
    sdk: &Sdk,
    writer: &mut W,
    processed_packages: &mut HashSet<UObject>,
    pkg: &ProcessedPackage,
) -> anyhow::Result<()> {
    let pkg_name = canonicalize_name(
        &pkg.package
            .get_full_name()
            .context("could not get package name")?,
    )
    .to_string();

    writeln!(writer, "pub mod {pkg_name};")?;

    let file = std::fs::File::create(format!("z:/tmp/ark_sdk/{pkg_name}.rs"))?;
    let mut file = BufWriter::new(file);
    generate_package_rust_module(pkg, sdk, &mut file)?;

    for (obj, _) in pkg.package_dependencies.iter() {
        if processed_packages.insert(*obj) {
            generate_package(
                sdk,
                writer,
                processed_packages,
                sdk.packages
                    .get(obj)
                    .context("could not find dependency package")?,
            )?;
        }
    }

    Ok(())
}

pub fn generate_sdk_to_tmp(sdk: &Sdk) -> anyhow::Result<()> {
    std::fs::create_dir_all("z:/tmp/ark_sdk/")?;
    let file = std::fs::File::create("z:/tmp/ark_sdk/mod.rs")?;
    let mut writer = BufWriter::new(file);

    writeln!(writer, "use core::cell::UnsafeCell;")?;
    writeln!(writer, "use core::ptr::NonNull;")?;
    writeln!(writer, "use super::AsPtr;")?;
    writeln!(writer, "use super::process_event;")?;
    writeln!(writer, "use super::StaticClass;")?;

    for (_, pkg) in &sdk.packages {
        let pkg_name = canonicalize_name(
            &pkg.package
                .get_full_name()
                .context("could not get package name")?,
        )
        .to_string();

        writeln!(writer, "pub mod {pkg_name};")?;

        let file = std::fs::File::create(format!("z:/tmp/ark_sdk/{pkg_name}.rs"))?;
        let mut writer = BufWriter::new(file);
        generate_package_rust_module(pkg, sdk, &mut writer)?;
    }

    Ok(())
}

pub fn generate_package_rust_module<W: Write>(
    pkg: &ProcessedPackage,
    sdk: &Sdk,
    w: &mut W,
) -> anyhow::Result<()> {
    writeln!(
        w,
        "#![allow(dead_code, unused_imports, non_snake_case, non_camel_case_types)]"
    )?;
    writeln!(w, "use super::*;")?;
    for (pkg, _) in &pkg.package_dependencies {
        writeln!(
            w,
            "use {}::*;",
            canonicalize_name(&pkg.get_full_name().context("could not get package name")?)
        )?;
    }

    for (_, ty) in pkg.types.iter().sorted_by(|a, b| {
        a.0.get_full_name_or_default()
            .cmp(&b.0.get_full_name_or_default())
    }) {
        match ty {
            Types::Class(class) => {
                generate_class(&class, sdk, w)?;
                generate_class_impl(&class, sdk, w)?;
            }
            Types::Enum(enm) => {
                generate_enum(enm, sdk, w)?;
            }
        }
    }
    Ok(())
}

pub(crate) fn inject_coreuobject_types(sdk: &mut Sdk) {
    let uobject = Class {
        is_class: true,
        size: 40,
        name: "UObject".to_string(),
        full_name: "Class CoreUObject.Object".to_string(),
        super_class: None,
        fields: vec![
            ClassField {
                offset: 0,
                size: 8,
                name: "vtbl".to_string(),
                ty: Type::Ptr(Box::new(Type::Primitive(PrimitiveType::U8))),
            },
            ClassField {
                offset: 8,
                size: 4,
                name: "object_flags".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("EObjectFlags")),
            },
            ClassField {
                offset: 12,
                size: 4,
                name: "internal_index".to_string(),
                ty: Type::Primitive(PrimitiveType::U32),
            },
            ClassField {
                offset: 16,
                size: 8,
                name: "class".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<UClass>")),
            },
            ClassField {
                offset: 24,
                size: 8,
                name: "name".to_string(),
                ty: Type::Name,
            },
            ClassField {
                offset: 32,
                size: 8,
                name: "outer".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<UObject>")),
            },
        ],
        methods: vec![],
    };

    let uobject = sdk.inject_type(uobject).expect("uobject");
    let ufield = Class {
        is_class: true,
        size: 48,
        name: "UField".to_string(),
        full_name: "Class CoreUObject.Field".to_string(),
        super_class: Some(unsafe { uobject.cast() }),
        fields: vec![ClassField {
            offset: 40,
            size: 8,
            name: "next".to_string(),
            ty: Type::Primitive(PrimitiveType::Custom("Option<UField>")),
        }],
        methods: vec![],
    };
    let ufield = sdk.inject_type(ufield).expect("ufield");

    let uenum = Class {
        is_class: true,
        size: 48,
        name: "UEnum".to_string(),
        full_name: "Class CoreUObject.Enum".to_string(),
        super_class: Some(unsafe { ufield.cast() }),
        fields: vec![
            ClassField {
                offset: 48,
                size: 16,
                name: "cpp_type".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("FString")),
            },
            ClassField {
                offset: 64,
                size: 16,
                name: "names".to_string(),
                ty: Type::Array(Box::new(Type::Name)),
            },
            ClassField {
                offset: 80,
                size: 4,
                name: "cpp_form_raw".to_string(),
                ty: Type::Primitive(PrimitiveType::U32),
            },
        ],
        methods: vec![],
    };
    let _uenum = sdk.inject_type(uenum).expect("uenum");

    let ustruct = Class {
        is_class: true,
        size: 144,
        name: "UStruct".to_string(),
        full_name: "Class CoreUObject.Struct".to_string(),
        super_class: Some(unsafe { ufield.cast() }),
        fields: vec![
            ClassField {
                offset: 48,
                size: 8,
                name: "super_struct".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<UStruct>")),
            },
            ClassField {
                offset: 56,
                size: 8,
                name: "children".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<UField>")),
            },
            ClassField {
                offset: 64,
                size: 4,
                name: "property_size".to_string(),
                ty: Type::Primitive(PrimitiveType::U32),
            },
            ClassField {
                offset: 68,
                size: 4,
                name: "min_alignment".to_string(),
                ty: Type::Primitive(PrimitiveType::U32),
            },
        ],
        methods: vec![],
    };
    let ustruct = sdk.inject_type(ustruct).expect("ustruct");

    let uproperty = Class {
        is_class: true,
        size: 144,
        name: "UProperty".to_string(),
        full_name: "Class CoreUObject.Property".to_string(),
        super_class: Some(unsafe { ufield.cast() }),
        fields: vec![
            ClassField {
                offset: 48,
                size: 4,
                name: "array_dimensions".to_string(),
                ty: Type::Primitive(PrimitiveType::I32),
            },
            ClassField {
                offset: 52,
                size: 4,
                name: "element_size".to_string(),
                ty: Type::Primitive(PrimitiveType::I32),
            },
            ClassField {
                offset: 56,
                size: 8,
                name: "property_flags".to_string(),
                ty: Type::Primitive(PrimitiveType::U64),
            },
            ClassField {
                offset: 64,
                size: 2,
                name: "rep_index".to_string(),
                ty: Type::Primitive(PrimitiveType::I16),
            },
            ClassField {
                offset: 68,
                size: 8,
                name: "rep_notify_function".to_string(),
                ty: Type::Name,
            },
            ClassField {
                offset: 76,
                size: 4,
                name: "offset".to_string(),
                ty: Type::Primitive(PrimitiveType::I32),
            },
        ],
        methods: vec![],
    };
    let _uproperty = sdk.inject_type(uproperty).expect("uproperty");

    let ufunction = Class {
        is_class: true,
        size: 176,
        name: "UFunction".to_string(),
        full_name: "Class CoreUObject.Function".to_string(),
        super_class: Some(unsafe { ustruct.cast() }),
        fields: vec![
            ClassField {
                offset: 144,
                size: 4,
                name: "function_flags".to_string(),
                ty: Type::Primitive(PrimitiveType::U32),
            },
            ClassField {
                offset: 148,
                size: 2,
                name: "rep_offset".to_string(),
                ty: Type::Primitive(PrimitiveType::U16),
            },
            ClassField {
                offset: 150,
                size: 1,
                name: "num_params".to_string(),
                ty: Type::Primitive(PrimitiveType::U8),
            },
            ClassField {
                offset: 152,
                size: 2,
                name: "params_size".to_string(),
                ty: Type::Primitive(PrimitiveType::U16),
            },
            ClassField {
                offset: 154,
                size: 2,
                name: "return_value_offset".to_string(),
                ty: Type::Primitive(PrimitiveType::U16),
            },
            ClassField {
                offset: 156,
                size: 2,
                name: "rpc_id".to_string(),
                ty: Type::Primitive(PrimitiveType::U16),
            },
            ClassField {
                offset: 158,
                size: 2,
                name: "rpc_response_id".to_string(),
                ty: Type::Primitive(PrimitiveType::U16),
            },
            ClassField {
                offset: 160,
                size: 8,
                name: "first_propert_to_init".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<UProperty>")),
            },
            ClassField {
                offset: 168,
                size: 8,
                name: "func".to_string(),
                ty: Type::Primitive(PrimitiveType::Custom("Option<NonNull<()>>")),
            },
        ],
        methods: vec![],
    };
    let _ufunction = sdk.inject_type(ufunction).expect("ufunction");
}
