//! This module is an attempt to define Unreal Types using just a pointer
//! because I thought that was proabably a lot safer and ergonomic than having a
//! bunch of nesting structures, and having to worry a lot more about aliasing
//! references. every type is also stored in an UnsafeCell which, if I
//! understand UnsafeCell and unsafe Rust correctly, should tell the compiler
//! that the contents might very well change under its feet (which they might).
#![allow(non_upper_case_globals)]
use bitflags::bitflags;
use std::{
    cell::UnsafeCell,
    fmt::{Debug, Display},
    hash::Hash,
    marker::PhantomData,
    ptr::NonNull,
};

pub mod any_type;

#[derive(Debug)]
#[repr(transparent)]
pub struct VTbl(NonNull<()>);

impl VTbl {}

bitflags! {
    #[repr(C)]
    pub struct EPropertyFlags: u64 {
        const None                              = 0x0000000000000000;
        const Edit                              = 0x0000000000000001;
        const ConstParm                         = 0x0000000000000002;
        const BlueprintVisible                  = 0x0000000000000004;
        const ExportObject                      = 0x0000000000000008;
        const BlueprintReadOnly                 = 0x0000000000000010;
        const Net                               = 0x0000000000000020;
        const EditFixedSize                     = 0x0000000000000040;
        const Parm                              = 0x0000000000000080;
        const OutParm                           = 0x0000000000000100;
        const ZeroConstructor                   = 0x0000000000000200;
        const ReturnParm                        = 0x0000000000000400;
        const DisableEditOnTemplate             = 0x0000000000000800;
        const Transient                         = 0x0000000000002000;
        const Config                            = 0x0000000000004000;
        const DisableEditOnInstance             = 0x0000000000010000;
        const EditConst                         = 0x0000000000020000;
        const GlobalConfig                      = 0x0000000000040000;
        const InstancedReference                = 0x0000000000080000;
        const DuplicateTransient                = 0x0000000000200000;
        const SaveGame                          = 0x0000000001000000;
        const NoClear                           = 0x0000000002000000;
        const ReferenceParm                     = 0x0000000008000000;
        const BlueprintAssignable               = 0x0000000010000000;
        const Deprecated                        = 0x0000000020000000;
        const IsPlainOldData                    = 0x0000000040000000;
        const RepSkip                           = 0x0000000080000000;
        const RepNotify                         = 0x0000000100000000;
        const Interp                            = 0x0000000200000000;
        const NonTransactional                  = 0x0000000400000000;
        const EditorOnly                        = 0x0000000800000000;
        const NoDestructor                      = 0x0000001000000000;
        const AutoWeak                          = 0x0000004000000000;
        const ContainsInstancedReference        = 0x0000008000000000;
        const AssetRegistrySearchable           = 0x0000010000000000;
        const SimpleDisplay                     = 0x0000020000000000;
        const AdvancedDisplay                   = 0x0000040000000000;
        const Protected                         = 0x0000080000000000;
        const BlueprintCallable                 = 0x0000100000000000;
        const BlueprintAuthorityOnly            = 0x0000200000000000;
        const TextExportTransient               = 0x0000400000000000;
        const NonPIEDuplicateTransient          = 0x0000800000000000;
        const ExposeOnSpawn                     = 0x0001000000000000;
        const PersistentInstance                = 0x0002000000000000;
        const UObjectWrapper                    = 0x0004000000000000;
        const HasGetValueTypeHash               = 0x0008000000000000;
        const NativeAccessSpecifierPublic       = 0x0010000000000000;
        const NativeAccessSpecifierProtected    = 0x0020000000000000;
        const NativeAccessSpecifierPrivate      = 0x0040000000000000;
        const SkipSerialization                 = 0x0080000000000000;
    }
}

bitflags! {
    #[repr(C)]
    pub struct EFunctionFlags: u32 {
        const None                      = 0x00000000;
        const Final                     = 0x00000001;
        const RequiredAPI               = 0x00000002;
        const BlueprintAuthorityOnly    = 0x00000004;
        const BlueprintCosmetic         = 0x00000008;
        const Net                       = 0x00000040;
        const NetReliable               = 0x00000080;
        const NetRequest                = 0x00000100;
        const Exec                      = 0x00000200;
        const Native                    = 0x00000400;
        const Event                     = 0x00000800;
        const NetResponse               = 0x00001000;
        const Static                    = 0x00002000;
        const NetMulticast              = 0x00004000;
        const UbergraphFunction         = 0x00008000;
        const MulticastDelegate         = 0x00010000;
        const Public                    = 0x00020000;
        const Private                   = 0x00040000;
        const Protected                 = 0x00080000;
        const Delegate                  = 0x00100000;
        const NetServer                 = 0x00200000;
        const HasOutParms               = 0x00400000;
        const HasDefaults               = 0x00800000;
        const NetClient                 = 0x01000000;
        const DLLImport                 = 0x02000000;
        const BlueprintCallable         = 0x04000000;
        const BlueprintEvent            = 0x08000000;
        const BlueprintPure             = 0x10000000;
        const EditorOnly                = 0x20000000;
        const Const                     = 0x40000000;
        const NetValidate               = 0x80000000;
        const AllFlags                  = 0xFFFFFFFF;
    }
}

bitflags! {
    #[repr(C)]
    pub struct EObjectFlags: u32 {
        const NoFlags                      = 0x00000000;
        const Public                       = 0x00000001;
        const Standalone                   = 0x00000002;
        const MarkAsNative                 = 0x00000004;
        const Transactional                = 0x00000008;
        const ClassDefaultObject           = 0x00000010;
        const ArchetypeObject              = 0x00000020;
        const Transient                    = 0x00000040;
        const MarkAsRootSet                = 0x00000080;
        const Unreachable                  = 0x00000100;
        const TagGarbageTemp               = 0x00000200;
        const NeedLoad                     = 0x00000400;
        const AsyncLoading                 = 0x00000800;
        const NeedPostLoad                 = 0x00001000;
        const NeedPostLoadSubobjects       = 0x00002000;
        const PendingKill                  = 0x00004000;
        const BeginDestroyed               = 0x00008000;
        const FinishDestroyed              = 0x00010000;
        const BeingRegenerated             = 0x00020000;
        const DefaultSubObject             = 0x00040000;
        const WasLoaded                    = 0x00080000;
        const TextExportTransient          = 0x00100000;
        const LoadCompleted                = 0x00200000;
        // not existant for 4.5.1 / ARK
        const InheritableComponentTemplate = 0x00400000;
        const DuplicateTransient           = 0x00800000;
        const StrongRefOnFrame             = 0x01000000;
        const NonPIEDuplicateTransient     = 0x02000000;
        const Dynamic                      = 0x04000000;
        const WillBeLoaded                 = 0x08000000;
        const HasExternalPackage           = 0x10000000;
        }
}

pub fn actor_static_class() -> Option<UClass> {
    use crate::global_tables::objects::{FindClass, GOBJECTS};
    use once_cell::sync::OnceCell;
    // this is something that we always do in C/C++ but unfortinately requires a blanket impl Sync + Send on all UObject types..
    static CLASS: OnceCell<Option<UClass>> = OnceCell::new();

    *CLASS.get_or_init(|| {
        let class = match GOBJECTS
            .read()
            .unwrap()
            .as_objects()
            .unwrap()
            .find_class("Class Engine.Actor")
        {
            Some(class) => Some(class),
            None => {
                log::error!("static class \"Class Engine.Actor\" not found!");
                None
            }
        };
        log::info!("class for {} is {:?}", "Class Engine.Actor", class);
        class
    })
}

macro_rules! define_utypes {
    ($($ty:ident),+) => {
        $(
            #[repr(transparent)]
            #[derive(Debug, Clone, Copy)]
            pub struct $ty(NonNull<UnsafeCell<()>>);

            impl_asuobject!($ty);

            impl const traits::UObject for $ty {}
        )+
    };
    ($($ty:ident => $name:literal),+) => {
        $(
            #[repr(transparent)]
            #[derive(Debug, Clone, Copy)]
            pub struct $ty(NonNull<UnsafeCell<()>>);

            impl traits::StaticClass for $ty {
                fn static_class_name() -> &'static str {
                    concat!("Class CoreUObject.", $name)
                }

                fn static_class() -> Option<UClass> {
                    use crate::global_tables::objects::{FindClass, GOBJECTS};
                    use once_cell::sync::OnceCell;
                    // this is something that we always do in C/C++ but unfortinately requires a blanket impl Sync + Send on all UObject types..
                    static CLASS: OnceCell<Option<UClass>> = OnceCell::new();

                    *CLASS.get_or_init(|| {
                        let class = match GOBJECTS
                            .read()
                            .unwrap()
                            .as_objects()
                            .unwrap()
                            .find_class(Self::static_class_name()) {
                                Some(class) => {Some(class)},
                                None => {
                                    log::error!("static class \"{}\" not found!", Self::static_class_name());
                                    None
                                }
                            };
                        log::info!("class for {} is {:?}", Self::static_class_name(), class);
                        class
                    })
                }
            }

            impl_asuobject!($ty);

            impl const traits::UObjectTrait for $ty {}
        )+
    };

    ($($ty:ident where $($trt:ty),+ => $name:literal),+) => {
        $(
            #[repr(transparent)]
            #[derive(Debug, Clone, Copy)]
            pub struct $ty(NonNull<UnsafeCell<()>>);

            impl traits::StaticClass for $ty {
                fn static_class_name() -> &'static str {
                    concat!("Class CoreUObject.", $name)
                }

                fn static_class() -> Option<UClass> {
                    use crate::global_tables::objects::{FindClass, GOBJECTS};
                    use once_cell::sync::OnceCell;
                    // this is something that we always do in C/C++ but unfortinately requires a blanket impl Sync + Send on all UObject types..
                    static CLASS: OnceCell<Option<UClass>> = OnceCell::new();

                    *CLASS.get_or_init(|| {
                        let class = match GOBJECTS
                            .read()
                            .unwrap()
                            .as_objects()
                            .unwrap()
                            .find_class(Self::static_class_name()) {
                                Some(class) => {Some(class)},
                                None => {
                                    log::error!("static class \"{}\" not found!", Self::static_class_name());
                                    None
                                }
                            };
                        log::info!("class for {} is {:?}", Self::static_class_name(), class);
                        class
                    })
                }
            }

            impl_asuobject!($ty);

            impl const traits::UObjectTrait for $ty {}
            $(
                impl const $trt for $ty {}
            )+
        )+
    };
}

macro_rules! impl_asuobject {
    ($($ty:ty),+) => {
        $(
            impl Eq for $ty {}
            impl<T> PartialEq<T> for $ty where T: AsUObject{
                fn eq(&self, other: &T) -> bool {
                    self.as_uobject().partial_eq(&other.as_uobject())
                }
            }

            unsafe impl Send for $ty {}
            unsafe impl Sync for $ty {}

        impl const traits::AsUObject for $ty {
            fn as_uobject(&self) -> self::UObject {
                self::UObject(self.0)
            }

            fn from_raw(raw: *mut ()) -> Option<Self> {
                match NonNull::new(raw) {
                    Some(ptr) => Some(Self(ptr.cast())),
                    None => None,
                }
            }

            fn from_nonnull(raw: NonNull<UnsafeCell<()>>) -> Self {
                Self(raw)
            }
        }
        )+
    };
}

define_utypes!(
    UObject => "Object"
);
define_utypes!(
    UField where UFieldTrait => "Field",
    UEnum where UFieldTrait, UEnumTrait => "Enum",
    UStruct where UFieldTrait, UStructTrait => "Struct",
    UScriptStruct where UFieldTrait, UStructTrait, UScriptStructTrait => "ScriptStruct",
    UClass where UFieldTrait, UStructTrait, UClassTrait => "Class",
    UProperty where UFieldTrait, UStructTrait, UPropertyTrait => "Property",
    UFunction where UFieldTrait, UStructTrait, UFunctionTrait => "Function",
    UNumericProperty where UFieldTrait, UStructTrait, UPropertyTrait, UNumericPropertyTrait => "NumericProperty",
    UByteProperty where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UBytePropertyTrait => "ByteProperty",
    UUInt16Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UUInt16PropertyTrait => "UInt16Property",
    UUInt32Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UUInt32PropertyTrait => "UInt32Property",
    UUInt64Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UUInt64PropertyTrait => "UInt64Property",
    UInt8Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UInt8PropertyTrait => "Int8Property",
    UInt16Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UInt16PropertyTrait => "Int16Property",
    UIntProperty where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UIntPropertyTrait => "IntProperty", // i32
    UInt64Property where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UInt64PropertyTrait => "Int64Property",
    UFloatProperty where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UFloatPropertyTrait => "FloatProperty",
    UDoubleProperty where UFieldTrait, UStructTrait, UPropertyTrait,UNumericPropertyTrait, UDoublePropertyTrait => "DoubleProperty",
    UBoolProperty where UFieldTrait, UStructTrait, UPropertyTrait,UBoolPropertyTrait => "BoolProperty",
    UObjectPropertyBase where UFieldTrait, UStructTrait, UPropertyTrait,UObjectPropertyBaseTrait => "ObjectPropertyBase",
    UObjectProperty where UFieldTrait, UStructTrait, UPropertyTrait, UObjectPropertyBaseTrait,UObjectPropertyTrait => "ObjectProperty",
    UClassProperty where UFieldTrait, UStructTrait, UPropertyTrait, UObjectPropertyBaseTrait, UObjectPropertyTrait,UClassPropertyTrait => "ClassProperty",
    UInterfaceProperty where UFieldTrait, UStructTrait, UPropertyTrait,UInterfacePropertyTrait => "InterfaceProperty",
    UWeakObjectProperty where UFieldTrait, UStructTrait, UPropertyTrait,UObjectPropertyBaseTrait, UObjectPropertyTrait,UWeakObjectPropertyTrait => "WeakObjectProperty",
    ULazyObjectProperty where UFieldTrait, UStructTrait, UPropertyTrait,UObjectPropertyBaseTrait, UObjectPropertyTrait,ULazyObjectPropertyTrait => "LazyObjectProperty",
    UAssetObjectProperty where UFieldTrait, UStructTrait, UPropertyTrait,UObjectPropertyBaseTrait, UObjectPropertyTrait,UAssetObjectPropertyTrait => "AssetObjectProperty",
    UAssetClassProperty where UFieldTrait, UStructTrait, UPropertyTrait,UObjectPropertyBaseTrait, UObjectPropertyTrait, UAssetObjectPropertyTrait,UAssetClassPropertyTrait => "AssetClassProperty",
    USoftObjectProperty where UFieldTrait, UStructTrait, UPropertyTrait, UObjectPropertyBaseTrait, USoftObjectPropertyTrait => "SoftObjectProperty",
    UNameProperty where UFieldTrait, UStructTrait, UPropertyTrait,UNamePropertyTrait => "NameProperty",
    UStructProperty where UFieldTrait, UStructTrait, UPropertyTrait,UStructPropertyTrait => "StructProperty",
    UStrProperty where UFieldTrait, UStructTrait, UPropertyTrait,UStrPropertyTrait => "StrProperty",
    UTextProperty where UFieldTrait, UStructTrait, UPropertyTrait,UTextPropertyTrait => "TextProperty",
    UArrayProperty where UFieldTrait, UStructTrait, UPropertyTrait,UArrayPropertyTrait => "ArrayProperty",
    UMapProperty where UFieldTrait, UStructTrait, UPropertyTrait,UMapPropertyTrait => "MapProperty",
    UDelegateProperty where UFieldTrait, UStructTrait, UPropertyTrait,UDelegatePropertyTrait => "DelegateProperty",
    UMulticastDelegateProperty where UFieldTrait, UStructTrait, UPropertyTrait,UMulticastDelegatePropertyTrait => "MulticastDelegateProperty",
    UEnumProperty where UFieldTrait, UStructTrait, UPropertyTrait,UEnumPropertyTrait => "EnumProperty"
);

impl Hash for UObject {
    fn hash<H: ~const std::hash::Hasher>(&self, state: &mut H) {
        self.0.hash(state);
    }
}

impl Display for UObject {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "UObject {{")?;
        writeln!(f, "\tvtbl: {:?}", self.vtbl())?;
        writeln!(f, "\tobject_flags{:?}", self.object_flags())?;
        writeln!(f, "\tinternal_index{:?}", self.internal_index())?;
        writeln!(f, "\tclass{:?}", self.class())?;
        writeln!(f, "\tname{:?}", &self.get_name_or_default())?;
        writeln!(f, "\touter{:?}", self.outer())?;
        writeln!(f, "}}")
    }
}

impl Display for UProperty {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "UProperty {{")?;
        writeln!(f, "name: {:?}", self.name().get_name())?;
        writeln!(f, "flags: {:?}", self.property_flags())?;
        writeln!(f, "offset: {:?}", self.offset())?;
        writeln!(f, "property_size: {:?}", self.property_size())?;
        writeln!(f, "element_size: {:?}", self.element_size())?;
        writeln!(f, "}}")?;

        Ok(())
    }
}

impl Display for UFunction {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "UFunction {{")?;
        writeln!(f, "name: {:?}", self.name().get_name())?;
        writeln!(f, "flags: {:?}", self.function_flags())?;
        writeln!(f, "num_args: {:?}", self.num_params())?;
        writeln!(f, "}}")?;

        Ok(())
    }
}

impl UObject {
    #![allow(dead_code)]

    pub fn inner(&self) -> NonNull<UnsafeCell<()>> {
        self.0
    }

    const fn raw_ptr(&self) -> *const u8 {
        unsafe { self.0.as_ref().get() as _ }
    }

    const fn raw_mut(&self) -> *mut u8 {
        unsafe { self.0.as_ref().get() as _ }
    }

    fn partial_eq(&self, other: &Self) -> bool {
        self.internal_index() == other.internal_index() && self.name() == other.name()
    }
}

#[derive(PartialEq, Eq, PartialOrd, Ord, Debug)]
pub enum ECppForm {
    Regular,
    Namespaced,
    EnumClass,
}

impl UEnum {
    pub fn cpp_form(&self) -> Option<ECppForm> {
        match *traits::UEnumTrait::cpp_form_raw(self) {
            0 => Some(ECppForm::Regular),
            1 => Some(ECppForm::Namespaced),
            2 => Some(ECppForm::EnumClass),
            _ => None,
        }
    }
}

pub struct OuterObjectIter<'a> {
    object: UObject,
    phantom: PhantomData<&'a ()>,
}

impl<'a> OuterObjectIter<'a> {
    pub fn new(object: UObject) -> Self {
        Self {
            object,
            phantom: PhantomData,
        }
    }
}

impl<'a> Iterator for OuterObjectIter<'a> {
    type Item = UObject;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(outer) = self.object.outer() {
            self.object = *outer;

            Some(self.object)
        } else {
            None
        }
    }
}

pub struct SuperStructIter<'a> {
    class: Option<UStruct>,
    phantom: PhantomData<&'a ()>,
}

impl<'a> SuperStructIter<'a> {
    pub fn new(class: UStruct) -> Self {
        Self {
            class: Some(class),
            phantom: PhantomData,
        }
    }
    pub fn from_option(class: Option<UStruct>) -> Self {
        Self {
            class,
            phantom: PhantomData,
        }
    }
}

impl<'a> Iterator for SuperStructIter<'a> {
    type Item = UStruct;

    fn next(&mut self) -> Option<Self::Item> {
        let next = *self.class?.super_field();

        core::mem::replace(&mut self.class, next)
    }
}

pub struct SuperClassIter<'a>(SuperStructIter<'a>);

impl<'a> SuperClassIter<'a> {
    pub fn new(class: UClass) -> Self {
        Self(SuperStructIter::new(unsafe { class.cast() }))
    }

    pub fn from_option(class: Option<UClass>) -> Self {
        Self(SuperStructIter::from_option(
            class.map(|c| unsafe { c.cast() }),
        ))
    }
}

impl<'a> Iterator for SuperClassIter<'a> {
    type Item = UClass;

    fn next(&mut self) -> Option<Self::Item> {
        self.0.next().map(|class| unsafe { class.cast() })
    }
}

pub struct ChildIter<'a> {
    children: Option<UField>,
    phantom: PhantomData<&'a ()>,
}

impl<'a> ChildIter<'a> {
    pub fn new(children: Option<UField>) -> Self {
        Self {
            children,
            phantom: PhantomData,
        }
    }
}

impl<'a> Iterator for ChildIter<'a> {
    type Item = UField;

    fn next(&mut self) -> Option<Self::Item> {
        let next = *self.children?.next();

        core::mem::replace(&mut self.children, next)
    }
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum UAnyType {
    UObject(UObject),
    UClass(UClass),
    UProperty(UProperty),
    UFunction(UFunction),
    UField(UField),
    UScriptStruct(UScriptStruct),
    UEnum(UEnum),
    UStruct(UStruct),
}

impl UAnyType {
    pub fn from_uobject(uobject: UObject) -> Self {
        // remember to order this from most specialized to least.
        // specifically: a UClass is also a UStruct, so check for UClass first,
        // then for any other subtypes of UStruct, then for UStruct
        if uobject.is_a_maybe(&UClass::static_class()) {
            Self::UClass(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UProperty::static_class()) {
            Self::UProperty(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UFunction::static_class()) {
            Self::UFunction(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UScriptStruct::static_class()) {
            Self::UScriptStruct(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UEnum::static_class()) {
            Self::UEnum(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UStruct::static_class()) {
            Self::UStruct(unsafe { uobject.cast() })
        } else if uobject.is_a_maybe(&UField::static_class()) {
            Self::UField(unsafe { uobject.cast() })
        } else {
            Self::UObject(uobject)
        }
    }
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum AnyNumericProp {
    U8(UByteProperty),
    U16(UUInt16Property),
    U32(UUInt32Property),
    U64(UUInt64Property),
    I8(UInt8Property),
    I16(UInt16Property),
    I32(UIntProperty),
    I64(UInt64Property),
    F32(UFloatProperty),
    F64(UDoubleProperty),
    Other(UNumericProperty),
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum AnyObjectProp {
    Object(UObjectProperty),
    Class(UClassProperty),
    Interface(UInterfaceProperty),
    WeakObject(UWeakObjectProperty),
    LazyObject(UWeakObjectProperty),
    SoftObject(UWeakObjectProperty),
    AssetObject(UAssetObjectProperty),
    AssetClass(UAssetClassProperty),
    Other(UObjectPropertyBase),
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum AnyDelegateProp {
    Delegate(UDelegateProperty),
    MulticastDelegate(UMulticastDelegateProperty),
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum AnyContainerProp {
    Array(UArrayProperty),
    Map(UMapProperty),
}

#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum AnyProp {
    Numeric(AnyNumericProp),
    Bool(UBoolProperty),
    ObjectBase(AnyObjectProp),
    Container(AnyContainerProp),
    Str(UStrProperty),
    Text(UTextProperty),
    Name(UNameProperty),
    Delegate(AnyDelegateProp),
    Enum(UEnumProperty),
    Struct(UStructProperty),
    Function(UFunction),
    Other(UProperty),
}

impl AnyNumericProp {
    fn from_uprop(prop: UNumericProperty) -> Self {
        if prop.is_a_maybe(&UByteProperty::static_class()) {
            Self::U8(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UUInt16Property::static_class()) {
            Self::U16(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UUInt32Property::static_class()) {
            Self::U32(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UUInt64Property::static_class()) {
            Self::U64(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UInt8Property::static_class()) {
            Self::I8(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UInt16Property::static_class()) {
            Self::I16(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UIntProperty::static_class()) {
            Self::I32(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UInt64Property::static_class()) {
            Self::I64(unsafe { prop.cast() })
        } else {
            Self::Other(prop)
        }
    }
}

impl AnyObjectProp {
    fn from_uprop(prop: UObjectPropertyBase) -> Self {
        if prop.is_a_maybe(&UInterfaceProperty::static_class()) {
            Self::Interface(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UAssetClassProperty::static_class()) {
            Self::AssetClass(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UAssetObjectProperty::static_class()) {
            Self::AssetObject(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UWeakObjectProperty::static_class()) {
            Self::WeakObject(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&ULazyObjectProperty::static_class()) {
            Self::LazyObject(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&USoftObjectProperty::static_class()) {
            Self::SoftObject(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UClassProperty::static_class()) {
            Self::Class(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UObjectProperty::static_class()) {
            Self::Object(unsafe { prop.cast() })
        } else {
            Self::Other(prop)
        }
    }
}

impl AnyProp {
    pub fn from_uprop(prop: UProperty) -> Self {
        if prop.is_a_maybe(&UNumericProperty::static_class()) {
            Self::Numeric(AnyNumericProp::from_uprop(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UBoolProperty::static_class()) {
            Self::Bool(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UObjectPropertyBase::static_class()) {
            Self::ObjectBase(AnyObjectProp::from_uprop(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UArrayProperty::static_class()) {
            Self::Container(AnyContainerProp::Array(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UMapProperty::static_class()) {
            Self::Container(AnyContainerProp::Map(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UDelegateProperty::static_class()) {
            Self::Delegate(AnyDelegateProp::Delegate(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UMulticastDelegateProperty::static_class()) {
            Self::Delegate(AnyDelegateProp::MulticastDelegate(unsafe { prop.cast() }))
        } else if prop.is_a_maybe(&UStrProperty::static_class()) {
            Self::Str(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UNameProperty::static_class()) {
            Self::Name(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UTextProperty::static_class()) {
            Self::Text(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UStructProperty::static_class()) {
            Self::Struct(unsafe { prop.cast() })
        } else if prop.is_a_maybe(&UEnumProperty::static_class()) {
            Self::Enum(unsafe { prop.cast() })
        } else {
            Self::Other(prop)
        }
    }
}

impl AsUObject for UAnyType {
    fn as_uobject(&self) -> self::UObject {
        match self {
            UAnyType::UObject(obj) => obj.as_uobject(),
            UAnyType::UClass(obj) => obj.as_uobject(),
            UAnyType::UField(obj) => obj.as_uobject(),
            UAnyType::UScriptStruct(obj) => obj.as_uobject(),
            UAnyType::UEnum(obj) => obj.as_uobject(),
            UAnyType::UStruct(obj) => obj.as_uobject(),
            UAnyType::UProperty(obj) => obj.as_uobject(),
            UAnyType::UFunction(obj) => obj.as_uobject(),
        }
    }

    fn from_raw(raw: *mut ()) -> Option<Self> {
        UObject::from_raw(raw).map(|obj| Self::from_uobject(obj))
    }

    fn from_nonnull(raw: NonNull<UnsafeCell<()>>) -> Self {
        Self::from_uobject(UObject::from_nonnull(raw))
    }
}

use traits::*;
pub mod traits {
    use std::cell::UnsafeCell;
    use std::ptr::NonNull;

    use anyhow::Context;
    use itertools::Itertools;

    use crate::fname::FName;
    use crate::tarray::{FString, TArray};

    use super::{EObjectFlags, VTbl};

    pub trait StaticClass {
        fn static_class_name() -> &'static str;
        fn static_class() -> Option<super::UClass>;
    }

    #[const_trait]
    pub trait AsUObject: Sized {
        fn as_uobject(&self) -> super::UObject;

        fn from_raw(raw: *mut ()) -> Option<Self>;

        fn from_nonnull(raw: NonNull<UnsafeCell<()>>) -> Self;

        unsafe fn cast<T>(&self) -> T
        where
            T: ~const AsUObject,
        {
            T::from_nonnull(self.as_uobject().0)
        }
    }

    pub trait UObjectNonConst: UObjectTrait {
        fn get_name(&self) -> anyhow::Result<String> {
            let name = self.name();
            name.get_name()
                .map(|name_str| {
                    if name.number > 0 {
                        format!("{}_{}", name_str, name.number)
                    } else {
                        name_str
                    }
                })
                .map(|name_str| {
                    name_str
                        .rfind("/")
                        .map(|pos| name_str[(pos + 1)..].to_string())
                        .unwrap_or_else(|| name_str)
                })
        }

        fn get_name_or_default(&self) -> String {
            self.get_name()
                .unwrap_or_else(|_| "DefaultObjectName".to_string())
        }

        fn get_full_name(&self) -> anyhow::Result<String> {
            let tmp = self
                .iter_outer_objects()
                .map(|obj| obj.get_name())
                .fold_ok(String::new(), |acc, obj_name| {
                    format!("{}.{}", obj_name, acc)
                })?;

            Ok(format!(
                "{} {}{}",
                self.class()
                    .context("invalid class pointer")?
                    .get_name_or_default(),
                tmp,
                self.get_name_or_default()
            ))
        }

        fn get_full_name_or_default(&self) -> String {
            self.get_full_name()
                .unwrap_or_else(|_| "DefaultObjectFullName".to_string())
        }

        fn iter_outer_objects(&self) -> super::OuterObjectIter {
            super::OuterObjectIter::new(self.as_uobject())
        }

        fn package_object(&self) -> super::UObject {
            self.iter_outer_objects()
                .last()
                .unwrap_or(self.as_uobject())
        }

        fn is_in_package(&self, package: super::UObject) -> bool {
            self.package_object() == package
        }

        fn is_a(&self, other: &super::UClass) -> bool {
            self.class()
                .map(|class| class.iter_super_classes().contains(other))
                .unwrap_or(false)
        }

        fn is_a_maybe(&self, other: &Option<super::UClass>) -> bool {
            other
                .as_ref()
                .map(|class| self.is_a(class))
                .unwrap_or(false)
        }
    }

    impl<T> UObjectNonConst for T where T: UObjectTrait {}

    #[const_trait]
    pub trait UObjectTrait: ~const AsUObject {
        fn vtbl(&self) -> &VTbl {
            unsafe { &*self.as_uobject().raw_ptr().offset(0).cast() }
        }
        fn object_flags(&self) -> &EObjectFlags {
            unsafe { &*self.as_uobject().raw_ptr().offset(8).cast() }
        }
        fn internal_index(&self) -> &u32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(12).cast() }
        }
        fn class(&self) -> &Option<super::UClass> {
            unsafe { &*self.as_uobject().raw_ptr().offset(16).cast() }
        }
        fn name(&self) -> &FName {
            unsafe { &*self.as_uobject().raw_ptr().offset(24).cast() }
        }
        fn outer(&self) -> &Option<super::UObject> {
            unsafe { &*self.as_uobject().raw_ptr().offset(32).cast() }
        }

        // utility

        fn is_package_object(&self) -> bool {
            self.outer().is_none()
        }
    }

    #[const_trait]
    pub trait UFieldTrait: ~const AsUObject {
        fn next(&self) -> &Option<super::UField> {
            unsafe { &*self.as_uobject().raw_ptr().offset(40).cast() }
        }
    }

    #[const_trait]
    pub trait UEnumTrait: ~const AsUObject {
        fn cpp_type(&self) -> &FString {
            unsafe { &*self.as_uobject().raw_ptr().offset(48).cast() }
        }
        fn names(&self) -> &TArray<FName> {
            unsafe { &*self.as_uobject().raw_ptr().offset(64).cast() }
        }
        fn cpp_form_raw(&self) -> &u32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(80).cast() }
        }
    }

    #[const_trait]
    pub trait UStructTrait: ~const AsUObject {
        fn super_field(&self) -> &Option<super::UStruct> {
            unsafe { &*self.as_uobject().raw_ptr().offset(48).cast() }
        }
        fn children(&self) -> &Option<super::UField> {
            unsafe { &*self.as_uobject().raw_ptr().offset(56).cast() }
        }
        fn property_size(&self) -> &u32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(64).cast() }
        }
        fn min_alignment(&self) -> &u32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(68).cast() }
        }
    }

    pub trait UStructNonConst: UStructTrait {
        fn iter_super_structs(&self) -> super::SuperStructIter {
            super::SuperStructIter::from_option(*self.super_field())
        }

        fn iter_children(&self) -> super::ChildIter {
            super::ChildIter::new(*self.children())
        }
    }

    impl<T> UStructNonConst for T where T: UStructTrait {}

    #[const_trait]
    pub trait UScriptStructTrait: ~const AsUObject {}

    pub trait UClassNonConst: UClassTrait {
        fn iter_super_classes(&self) -> super::SuperClassIter {
            super::SuperClassIter::new(unsafe { self.cast() })
        }
    }

    impl<T> UClassNonConst for T where T: UClassTrait {}

    #[const_trait]
    pub trait UClassTrait: ~const UStructTrait {}

    #[const_trait]
    pub trait UFunctionTrait: ~const AsUObject {
        fn function_flags(&self) -> &super::EFunctionFlags {
            unsafe { &*self.as_uobject().raw_ptr().offset(144).cast() }
        }
        fn rep_offset(&self) -> &u16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(148).cast() }
        }
        fn num_params(&self) -> &u8 {
            unsafe { &*self.as_uobject().raw_ptr().offset(150).cast() }
        }
        fn params_size(&self) -> &u16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(152).cast() }
        }
        fn return_value_offset(&self) -> &u16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(154).cast() }
        }
        fn rpc_id(&self) -> &u16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(156).cast() }
        }
        fn rpc_response_id(&self) -> &u16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(158).cast() }
        }
        fn first_property_to_init(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(160).cast() }
        }
        fn func(&self) -> &Option<NonNull<()>> {
            unsafe { &*self.as_uobject().raw_ptr().offset(168).cast() }
        }
    }

    #[const_trait]
    pub trait UPropertyTrait: ~const AsUObject {
        fn array_dim(&self) -> &i32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(48).cast() }
        }
        fn element_size(&self) -> &i32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(52).cast() }
        }
        fn property_flags(&self) -> &super::EPropertyFlags {
            unsafe { &*self.as_uobject().raw_ptr().offset(56).cast() }
        }
        fn rep_index(&self) -> &i16 {
            unsafe { &*self.as_uobject().raw_ptr().offset(64).cast() }
        }
        fn rep_notify_function(&self) -> &FName {
            unsafe { &*self.as_uobject().raw_ptr().offset(68).cast() }
        }
        fn offset(&self) -> &i32 {
            unsafe { &*self.as_uobject().raw_ptr().offset(76).cast() }
        }
        fn property_link_next(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(80).cast() }
        }
        fn next_ref(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(88).cast() }
        }
        fn destructor_link_next(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(96).cast() }
        }
        fn post_construct_link_next(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(104).cast() }
        }
    }

    #[const_trait]
    pub trait UNamePropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UStrPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UTextPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UNumericPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UUInt16PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UUInt32PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UUInt64PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UInt16PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UInt8PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UInt64PropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UIntPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UFloatPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UDoublePropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UBytePropertyTrait: ~const AsUObject {
        fn uenum(&self) -> &Option<super::UEnum> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UBoolPropertyTrait: ~const AsUObject {
        fn field_size(&self) -> &u8 {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
        fn byte_offset(&self) -> &u8 {
            unsafe { &*self.as_uobject().raw_ptr().offset(113).cast() }
        }
        fn byte_mask(&self) -> &u8 {
            unsafe { &*self.as_uobject().raw_ptr().offset(114).cast() }
        }
        fn field_mask(&self) -> &u8 {
            unsafe { &*self.as_uobject().raw_ptr().offset(115).cast() }
        }
    }

    #[const_trait]
    pub trait UObjectPropertyBaseTrait: ~const AsUObject {
        fn property_class(&self) -> &Option<super::UClass> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UObjectPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UWeakObjectPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait ULazyObjectPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UAssetObjectPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait USoftObjectPropertyTrait: ~const AsUObject {}

    #[const_trait]
    pub trait UInterfacePropertyTrait: ~const AsUObject {
        fn interface_class(&self) -> &Option<super::UClass> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UStructPropertyTrait: ~const AsUObject {
        fn ustruct(&self) -> &Option<super::UStruct> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UArrayPropertyTrait: ~const AsUObject {
        fn inner(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UMapPropertyTrait: ~const AsUObject {
        fn key_prop(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
        fn value_prop(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(120).cast() }
        }
    }

    #[const_trait]
    pub trait UClassPropertyTrait: ~const AsUObject {
        fn meta_class(&self) -> &Option<super::UClass> {
            unsafe { &*self.as_uobject().raw_ptr().offset(120).cast() }
        }
    }

    #[const_trait]
    pub trait UAssetClassPropertyTrait: ~const AsUObject {
        fn meta_class(&self) -> &Option<super::UClass> {
            unsafe { &*self.as_uobject().raw_ptr().offset(120).cast() }
        }
    }

    #[const_trait]
    pub trait UDelegatePropertyTrait: ~const AsUObject {
        fn signature_function(&self) -> &Option<super::UFunction> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UMulticastDelegatePropertyTrait: ~const AsUObject {
        fn signature_function(&self) -> &Option<super::UFunction> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
    }

    #[const_trait]
    pub trait UEnumPropertyTrait: ~const AsUObject {
        fn underlying_type(&self) -> &Option<super::UProperty> {
            unsafe { &*self.as_uobject().raw_ptr().offset(112).cast() }
        }
        fn uenum(&self) -> &Option<super::UEnum> {
            unsafe { &*self.as_uobject().raw_ptr().offset(120).cast() }
        }
    }
}
